#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import os
import shutil
import subprocess
import sys

import ait.backend.xilinx.utils.checkers as checkers
from ait.frontend.utils import msg

script_folder = os.path.basename(os.path.dirname(os.path.realpath(__file__)))


def run_step(project_args):
    global start_time
    global project_backend_path
    global petalinux_build_path
    global petalinux_install_path

    start_time = project_args['start_time']
    project_path = project_args['path']
    board = project_args['board']
    args = project_args['args']

    project_backend_path = project_path + '/' + args.backend

    petalinux_build_path = os.path.realpath(os.getenv('PETALINUX_BUILD')) if os.getenv('PETALINUX_BUILD') else ''
    petalinux_install_path = os.path.realpath(os.getenv('PETALINUX_INSTALL')) if os.getenv('PETALINUX_INSTALL') else ''

    checkers.check_petalinux(petalinux_build_path, petalinux_install_path)

    path_hdf = project_backend_path + '/' + args.name + '/' + args.name + '.sdk/'

    if os.path.exists(petalinux_install_path + '/.version-history'):
        # Seems to be petalinux 2019.1 or later (may match other untested versions)
        command = 'petalinux-config --silentconfig --get-hw-description=' + path_hdf
    else:
        # Seems to be petalinux 2018.3 or previous (may match other untested versions)
        command = 'petalinux-config --oldconfig --get-hw-description=' + path_hdf

    if args.verbose_info:
        msg.log('> ' + command)
    p = subprocess.Popen(command, stdout=sys.stdout.subprocess, stderr=sys.stdout.subprocess,
                         cwd=petalinux_build_path, shell=True)

    if args.verbose:
        for line in iter(p.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))

    retval = p.wait()
    if retval:
        msg.error('Generation of petalinux boot files failed', start_time, False)

    if os.path.exists(petalinux_build_path + '/subsystems/linux/configs/device-tree/'):
        # Seems to be petalinux 2016.3 (may match other untested versions)
        if args.verbose_info:
            msg.log('Fixing devicetree (2016.3 mode)')

        petalinux_build_dts_path = petalinux_build_path + '/subsystems/linux/configs/device-tree/'
        shutil.copy2(project_backend_path + '/' + args.name + '/pl_ompss_at_fpga.dtsi', petalinux_build_dts_path)

        content_dtsi = None
        with open(petalinux_build_dts_path + '/system-conf.dtsi', 'r') as file:
            content_dtsi = file.read().splitlines()

        line = [idx for idx in range(len(content_dtsi)) if content_dtsi[idx].find('/include/ "pl.dtsi"') != -1]
        content_dtsi.insert(line[0] + 1, '/include/ \"pl_ompss_at_fpga.dtsi\"')

        board_dtsi_fix_file = project_backend_path + '/board/' + board.name + '/' + board.name + '_boot.dtsi'
        if os.path.exists(board_dtsi_fix_file):
            shutil.copy2(board_dtsi_fix_file, petalinux_build_dts_path)
            content_dtsi.insert(line[0] + 2, '/include/ \"' + board.name + '_boot.dtsi\"')

        with open(petalinux_build_dts_path + '/system-conf.dtsi', 'w') as file:
            file.write('\n'.join(content_dtsi))

        command = 'petalinux-build -c bootloader -x mrproper'
        if args.verbose_info:
            msg.log('> ' + command)
        p = subprocess.Popen(command, stdout=sys.stdout.subprocess, stderr=sys.stdout.subprocess,
                             cwd=petalinux_build_path, shell=True)
        if args.verbose:
            for line in iter(p.stdout.readline, b''):
                sys.stdout.write(line.decode('utf-8'))

        retval = p.wait()
        if retval:
            msg.error('Generation of petalinux boot files failed', start_time, False)
    elif os.path.exists(petalinux_build_path + '/project-spec/meta-user/recipes-bsp/device-tree/files/'):
        # Seems to be petalinux 2018.3 or 2019.1 (may match other untested versions)
        if args.verbose_info:
            msg.log('Fixing devicetree (2018.3 mode)')

        petalinux_build_dts_path = petalinux_build_path + '/project-spec/meta-user/recipes-bsp/device-tree/files/'

        content_dtsi = None
        with open(petalinux_build_dts_path + '/system-user.dtsi', 'r') as file:
            content_dtsi = file.read().splitlines()

        # Remove old includes to pl_bsc.dtsi and insert the new one
        line = [idx for idx in range(len(content_dtsi)) if content_dtsi[idx].find('pl_bsc.dtsi') != -1]
        if len(line) == 1:
            content_dtsi.pop(line[0])
        elif len(line) > 1:
            msg.error('Uncontrolled path in run_boot_step: more than 1 line of system-user.dtsi contains pl_bsc.dtsi')

        # Remove old includes to pl_ompss_at_fpga.dtsi and insert the new one
        line = [idx for idx in range(len(content_dtsi)) if content_dtsi[idx].find('pl_ompss_at_fpga.dtsi') != -1]
        if len(line) == 1:
            content_dtsi.pop(line[0])
        elif len(line) > 1:
            msg.error('Uncontrolled path in run_boot_step: more than 1 line of system-user.dtsi contains pl_ompss_at_fpga.dtsi')

        line = [idx for idx in range(len(content_dtsi)) if content_dtsi[idx].find('pl_ompss_at_fpga.dtsi') != -1]
        content_dtsi.insert(len(content_dtsi), '/include/ \"' + project_backend_path + '/' + args.name + '/pl_ompss_at_fpga.dtsi' + '\"')

        # Remove old includes to <board>_boot.dtsi and insert the new one
        line = [idx for idx in range(len(content_dtsi)) if content_dtsi[idx].find(board.name + '_boot.dtsi') != -1]
        if len(line) == 1:
            content_dtsi.pop(line[0])
        elif len(line) > 1:
            msg.error('Uncontrolled path in run_boot_step: more than 1 line of system-user.dtsi contains <board>_bsc.dtsi')

        board_dtsi_fix_file = project_backend_path + '/board/' + board.name + '/' + board.name + '_boot.dtsi'
        if os.path.exists(board_dtsi_fix_file):
            shutil.copy2(board_dtsi_fix_file, petalinux_build_dts_path)
            content_dtsi.insert(len(content_dtsi), '/include/ \"' + board_dtsi_fix_file + '\"')

        with open(petalinux_build_dts_path + '/system-user.dtsi', 'w') as file:
            file.write('\n'.join(content_dtsi))
    else:
        msg.warning('Devicetree fix failed. Petalinux version cannot be determined. Continuing anyway...')

    command = 'petalinux-build'
    if args.verbose_info:
        msg.log('> ' + command)
    p = subprocess.Popen(command, stdout=sys.stdout.subprocess, stderr=sys.stdout.subprocess,
                         cwd=petalinux_build_path, shell=True)
    if args.verbose:
        for line in iter(p.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))

    retval = p.wait()
    if retval:
        msg.error('Generation of petalinux boot files failed', start_time, False)

    path_bit = project_path + '/' + args.name + '.bit'
    command = 'petalinux-package --force --boot --fsbl ./images/linux/*_fsbl.elf'
    command += ' --fpga ' + path_bit + ' --u-boot ./images/linux/u-boot.elf'
    if args.verbose_info:
        msg.log('> ' + command)
    p = subprocess.Popen(command, stdout=sys.stdout.subprocess, stderr=sys.stdout.subprocess,
                         cwd=petalinux_build_path, shell=True)
    if args.verbose:
        for line in iter(p.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))

    retval = p.wait()
    if retval:
        msg.error('Generation of petalinux boot files failed', start_time, False)
    else:
        shutil.copy2(petalinux_build_path + '/images/linux/BOOT.BIN', project_path)
        shutil.copy2(petalinux_build_path + '/images/linux/image.ub', project_path)
        msg.success('Petalinux boot files generated')



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #


import importlib
import math
import os
import re
import sys
import time

import setuptools


class Accelerator:
    def __init__(self, acc_config):
        for attribute in acc_config:
            setattr(self, attribute, acc_config[attribute])


class Color:
    GREEN = '\033[0;32m'   # Success
    CYAN = '\033[0;36m'    # Info
    YELLOW = '\033[0;33m'  # Warning
    RED = '\033[1;31m'     # Error
    END = '\033[0m'


class Messages:
    def __getHeader(self):
        header = '[AIT] ' if self.verbose else ''
        header += time.strftime('[%Y-%m-%d %H:%M] ', time.localtime()) if self.__printtime else ''
        return header

    def __init__(self):
        self.name = False
        self.__printtime = False
        self.verbose = False

    def setProjectName(self, name):
        self.name = name

    def setPrintTime(self, mode):
        self.__printtime = mode

    def setVerbose(self, verbose):
        self.verbose = verbose

    def success(self, msg):
        print(self.__getHeader() + Color.GREEN + msg + Color.END)

    def info(self, msg):
        print(self.__getHeader() + Color.CYAN + 'INFO: ' + msg + Color.END)

    def warning(self, msg):
        print(self.__getHeader() + Color.YELLOW + 'WARNING: ' + msg + Color.END)

    def error(self, msg, start_time=None, simple=True):
        if self.name and not simple:
            print(self.__getHeader() + Color.RED + 'ERROR: ' + msg + ' after ' + secondsToHumanReadable(int(time.time() - start_time)) + '. Check ' + self.name + '.ait.log for more information' + Color.END)
        else:
            print(self.__getHeader() + Color.RED + 'ERROR: ' + msg + Color.END)
        sys.exit(1)

    def log(self, msg):
        print(self.__getHeader() + msg)


def decimalToHumanReadable(number, precision=0):
    log_number = math.log2(number)

    # Remove decimal point when 0 decimals of precision
    if precision == 0:
        precision -= 1

    # Using 15 decimals of precision to avoid rounding
    if log_number > 30:
        return '{:.15f}'.format(number / 1024**3)[:-15 + precision] + 'G'
    elif log_number > 20:
        return '{:.15f}'.format(number / 1024**2)[:-15 + precision] + 'M'
    elif log_number > 10:
        return '{:.15f}'.format(number / 1024)[:-15 + precision] + 'K'
    else:
        return '{:.15f}'.format(number)[:-15 + precision]


def decimalFromHumanReadable(number):
    if re.match(r'[a-zA-Z]', number[-1:]):
        if (number[:-1]).replace('.', '', 1).isdigit() and float(number[:-1]) > 0:
            if number[-1:] == "G":
                return int(float(number[:-1]) * 1024**3)
            elif number[-1:] == "M":
                return int(float(number[:-1]) * 1024**2)
            elif number[-1:] == "K":
                return int(float(number[:-1]) * 1024)
            else:
                raise TypeError('invalid unit')
        else:
            raise ValueError('invalid value')

    else:
        if number.isdigit() and int(number) > 0:
            return int(number)
        else:
            raise ValueError('invalid value')


def secondsToHumanReadable(seconds):
    TIME_DURATION_UNITS = (
        ('week', 60 * 60 * 24 * 7),
        ('day', 60 * 60 * 24),
        ('hour', 60 * 60),
        ('min', 60),
        ('sec', 1)
    )

    if seconds == 0:
        return 'inf'
    parts = []
    for unit, div in TIME_DURATION_UNITS:
        amount, seconds = divmod(int(seconds), div)
        if amount > 0:
            parts.append('{} {}{}'.format(amount, unit, '' if amount == 1 else 's'))
    return ', '.join(parts)


ait_path = os.path.normpath(os.path.dirname(os.path.realpath(__file__)) + '/..')

backends = dict()
for backend_subpkg in setuptools.find_packages(where=ait_path + '/backend', exclude=['*.*']):
    backend = importlib.import_module('ait.backend.%s.info' % (backend_subpkg))
    backends[backend.NAME] = backend.info

msg = Messages()



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import glob
import os
import re
import shutil
import subprocess
import sys

import ait.backend.xilinx.utils.checkers as checkers
from ait.frontend.utils import ait_path, msg
from ait.backend.xilinx.utils.checkers import vivado_version

script_folder = os.path.basename(os.path.dirname(os.path.realpath(__file__)))


def gen_utilization_report(out_path):
    used_resources = {}
    av_resources = {}
    util_resources = {}

    # Check implementation reports path
    rpt_path = project_backend_path + '/' + args.name + '/' + args.name + '.runs/impl_1'
    rpt_path += '/' + args.name + '_design_wrapper_utilization_placed.rpt'
    if not os.path.exists(rpt_path):
        msg.warning('Cannot find rpt file. Skipping bitstream utilization report')
        return

    with open(rpt_path, 'r') as rpt_file:
        rpt_data = rpt_file.readlines()

        if vivado_version >= '2020.2':
            used_offset = 2
            av_offset = 5
            util_offset = 6
        else:
            used_offset = 2
            av_offset = 4
            util_offset = 5

        # Search LUT/FF section
        # NOTE: Possible section names: Slice Logic, CLB Logic
        ids = [idx for idx in range(len(rpt_data) - 1) if ((re.match(r'^[0-9]\. ' + 'Slice Logic\n', rpt_data[idx])
                                                           and rpt_data[idx + 1] == '--------------\n')
                                                           or (re.match(r'^[0-9]\. ' + 'CLB Logic\n', rpt_data[idx])
                                                           and rpt_data[idx + 1] == '------------\n'))]
        if len(ids) != 1:
            msg.warning('Cannot find LUT/FF info in rpt file. Skipping bitstream utilization report')
            return

        # Get LUT
        elems = rpt_data[ids[0] + 6].split('|')
        used_resources['LUT'] = elems[used_offset].strip()
        av_resources['LUT'] = elems[av_offset].strip()
        util_resources['LUT'] = elems[util_offset].strip()

        # Get FF
        elems = rpt_data[ids[0] + 11].split('|')
        used_resources['FF'] = elems[used_offset].strip()
        av_resources['FF'] = elems[av_offset].strip()
        util_resources['FF'] = elems[util_offset].strip()

        # Get DSP
        # NOTE: Possible section names: DSP, ARITHMETIC
        ids = [idx for idx in range(len(rpt_data) - 1) if ((re.match(r'^[0-9]\. ' + 'DSP\n', rpt_data[idx])
                                                           and rpt_data[idx + 1] == '------\n')
                                                           or (re.match(r'^[0-9]\. ' + 'ARITHMETIC\n', rpt_data[idx])
                                                           and rpt_data[idx + 1] == '-------------\n'))]
        if len(ids) != 1:
            msg.warning('Cannot find DSP info in rpt file. Skipping bitstream utilization report')
            return
        elems = rpt_data[ids[0] + 6].split('|')
        used_resources['DSP'] = elems[used_offset].strip()
        av_resources['DSP'] = elems[av_offset].strip()
        util_resources['DSP'] = elems[util_offset].strip()

        # Search BRAM/URAM
        # NOTE: Possible section names: Memory, BLOCKRAM
        ids = [idx for idx in range(len(rpt_data) - 1) if ((re.match(r'^[0-9]\. ' + 'Memory\n', rpt_data[idx])
                                                           and rpt_data[idx + 1] == '---------\n')
                                                           or (re.match(r'^[0-9]\. ' + 'BLOCKRAM\n', rpt_data[idx])
                                                           and rpt_data[idx + 1] == '-----------\n'))]
        if len(ids) != 1:
            msg.warning('Cannot find BRAM info in rpt file. Skipping bitstream utilization report')
            return

        # BRAM
        elems = rpt_data[ids[0] + 6].split('|')
        used_resources['BRAM'] = str(int(float(elems[used_offset].strip()) * 2))
        av_resources['BRAM'] = str(int(float(elems[av_offset].strip()) * 2))
        util_resources['BRAM'] = elems[util_offset].strip()

        # URAM
        # NOTE: It is not placed in the same offset for all boards (search in some lines)
        # NOTE: It is not available in all boards, so check if valid data is found
        ids = [idx for idx in range(ids[0] + 6, ids[0] + 20) if ((re.match(r'^| URAM', rpt_data[idx])))]
        for idx in ids:
            elems = rpt_data[idx].split('|')
            if len(elems) >= 6 and elems[1].strip() == 'URAM':
                used_resources['URAM'] = elems[used_offset].strip()
                av_resources['URAM'] = elems[av_offset].strip()
                util_resources['URAM'] = elems[util_offset].strip()
                break

    resources_file = open(out_path, 'w')
    msg.log('Resources utilization summary')
    for name in ['BRAM', 'DSP', 'FF', 'LUT', 'URAM']:
        # Check if resource is available
        if name not in used_resources:
            continue

        report_string = '{0:<9} {1:>6} used | {2:>6} available - {3:>6}% utilization'
        report_string_formatted = report_string.format(name, used_resources[name],
                                                       av_resources[name], util_resources[name])
        msg.log(report_string_formatted)
        resources_file.write(report_string_formatted + '\n')
    resources_file.close()


def gen_wns_report(out_path):
    wns = None
    tns = None
    num_fail = 0
    num_total = 0

    # Check implementation reports path
    rpt_path = project_backend_path + '/' + args.name + '/' + args.name + '.runs/impl_1'
    rpt_path += '/' + args.name + '_design_wrapper_timing_summary_routed.rpt'
    if not os.path.exists(rpt_path):
        msg.warning('Cannot find rpt file. Skipping WNS report')
        return

    with open(rpt_path, 'r') as rpt_file:
        rpt_data = rpt_file.readlines()

        # Search header line
        ids = [idx for idx in range(len(rpt_data) - 1) if (re.match(r'^\s+WNS\(ns\)\s+TNS\(ns\)\s+', rpt_data[idx]))]
        if len(ids) != 1:
            msg.warning('Cannot find WNS report table header. Skipping WNS report')
            return

        # Get information from 1st row
        elems = rpt_data[ids[0] + 2].split()
        wns = float(elems[0])
        tns = float(elems[1])
        num_fail = int(elems[2])
        num_total = int(elems[3])

    msg.log('Worst Negative Slack (WNS) summary')
    if wns >= 0.0:
        msg.success(str(num_fail) + ' endpoints of ' + str(num_total) + ' have negative slack (WNS: '
                    + str(wns) + ')')
    else:
        msg.warning(str(num_fail) + ' endpoints of ' + str(num_total) + ' have negative slack (WNS: '
                    + str(wns) + ', TNS: ' + str(tns) + ')')

    with open(out_path, 'w') as timing_file:
        timing_file.write('WNS ' + str(wns) + '\n')
        timing_file.write('TNS ' + str(tns) + '\n')
        timing_file.write('NUM_ENDPOINTS ' + str(num_total) + '\n')
        timing_file.write('NUM_FAIL_ENDPOINTS ' + str(num_fail))


def run_step(project_args):
    global args
    global board
    global chip_part
    global start_time
    global ait_backend_path
    global project_backend_path

    args = project_args['args']
    board = project_args['board']
    start_time = project_args['start_time']
    project_path = project_args['path']

    chip_part = board.chip_part + ('-' + board.es if (board.es and not args.ignore_eng_sample) else '')
    ait_backend_path = ait_path + '/backend/' + args.backend
    project_backend_path = project_path + '/' + args.backend

    # Check if Vivado requirements are met
    checkers.check_vivado()

    if os.path.isfile(project_backend_path + '/' + args.name + '/' + args.name + '.xpr'):
        # Enable beta device on Vivado init script
        if board.board_part:
            p = subprocess.Popen('echo "enable_beta_device ' + chip_part + '\nset_param board.repoPaths [list '
                                 + project_backend_path + '/board/' + board.name + '/board_files]" > '
                                 + project_backend_path + '/vivado.tcl', shell=True)
            retval = p.wait()
        else:
            p = subprocess.Popen('echo "enable_beta_device ' + chip_part + '" > '
                                 + project_backend_path + '/vivado.tcl', shell=True)
            retval = p.wait()

        p = subprocess.Popen('vivado -init -nojournal -nolog -notrace -mode batch -source '
                             + project_backend_path + '/tcl/scripts/generate_bitstream.tcl',
                             cwd=project_backend_path,
                             stdout=sys.stdout.subprocess,
                             stderr=sys.stdout.subprocess, shell=True)

        if args.verbose:
            for line in iter(p.stdout.readline, b''):
                sys.stdout.write(line.decode('utf-8'))

        retval = p.wait()
        if retval:
            msg.error('Bitstream generation failed', start_time, False)
        else:
            if board.arch.device == 'zynq' and checkers.check_bootgen():
                bif_file = open(project_backend_path + '/' + args.name + '/' + args.name + '.runs/impl_1/bitstream.bif', 'w')
                bif_file.write('all:\n'
                               + '{\n'
                               + '\t' + args.name + '_design_wrapper.bit\n'
                               + '}')
                bif_file.close()
                p = subprocess.Popen('bootgen -image bitstream.bif -arch zynq -process_bitstream bin -w',
                                     cwd=project_backend_path + '/' + args.name + '/' + args.name + '.runs/impl_1',
                                     stdout=sys.stdout.subprocess,
                                     stderr=sys.stdout.subprocess, shell=True)

                if args.verbose:
                    for line in iter(p.stdout.readline, b''):
                        sys.stdout.write(line.decode('utf-8'))

                retval = p.wait()
                if retval:
                    msg.warning('Could not create .bit.bin file')
                else:
                    shutil.copy2(glob.glob(project_backend_path + '/' + args.name + '/' + args.name
                                 + '.runs/impl_1/' + args.name + '*.bit.bin')[0],
                                 project_path + '/' + args.name + '.bit.bin')

            shutil.copy2(glob.glob(project_backend_path + '/' + args.name + '/' + args.name
                         + '.runs/impl_1/' + args.name + '*.bit')[0],
                         project_path + '/' + args.name + '.bit')
            if board.arch.device == 'zynq' or board.arch.device == 'zynqmp':
                shutil.copy2(glob.glob(project_backend_path + '/' + args.name + '/' + args.name
                             + '.runs/impl_1/' + args.name + '*.bin')[0],
                             project_path + '/' + args.name + '.bin')
            gen_utilization_report(project_path + '/' + args.name + '.resources-impl.txt')
            gen_wns_report(project_path + '/' + args.name + '.timing-impl.txt')
            msg.success('Bitstream generated')
    else:
        msg.error('No Vivado .xpr file exists for the current project. Bitstream generation failed')



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import argparse
import json
import math
import os

from ait.frontend.utils import decimalFromHumanReadable, decimalToHumanReadable, msg


class StoreChoiceValue(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.choices.index(values))


# Custom argparse type representing a path to a file
class FileType:
    def __call__(self, arg):
        if os.path.isfile(arg):
            return os.path.realpath(arg)
        else:
            raise argparse.ArgumentTypeError('Invalid file')


# Custom argparse type representing a bounded int
class IntRangeType:
    def __init__(self, imin=None, imax=None):
        self.imin = imin
        self.imax = imax

    def __call__(self, arg):
        try:
            value = int(arg)
        except ValueError:
            raise self.exception()
        if (self.imin is not None and value < self.imin) or (self.imax is not None and value > self.imax):
            raise self.exception()
        return value

    def exception(self):
        if self.imin is not None and self.imax is not None:
            return argparse.ArgumentTypeError('must be an integer in the range [{}, {}]'.format(self.imin, self.imax))
        elif self.imin is not None:
            return argparse.ArgumentTypeError('must be an integer >= {}'.format(self.imin))
        elif self.imax is not None:
            return argparse.ArgumentTypeError('must be an integer <= {}'.format(self.imax))
        else:
            return argparse.ArgumentTypeError('must be an integer')


class ArgParser():
    def parse_known_args(self, args=None, namespace=None):
        # Check if configuration file exists and parse its values
        config_file_path = os.path.expanduser('~') + '/.ait/config.json'
        if os.path.exists(config_file_path):
            with open(config_file_path) as config_file:
                config = json.load(config_file)
            self.parser.set_defaults(**config)

        # Parse values from argv
        args, extras = self.parser.parse_known_args(args, namespace)
        self.check_args(args)

        return args, extras

    def is_default(self, dest, backend):
        value = False
        if self.defaults[backend][dest] is not None:
            value = self.defaults[backend][dest]['used']
        return value

    def __init__(self):
        # Create vendor-specific parser
        # NOTE: usage must be suppressed as this parser is not called directly
        self.parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, epilog='  environment variables:\n    PETALINUX_INSTALL\tpath where Petalinux is installed\n    PETALINUX_BUILD\tpath where the Petalinux project is located', prog='ait', usage=argparse.SUPPRESS)

        # Rename help title
        self.parser._optionals.title = 'Xilinx-specific arguments'

        # Vendor-specific arguments
        self.parser.add_argument('--floorplanning_constr', help='built-in floorplanning constraints for accelerators and static logic\nacc: accelerator kernels are constrained to a SLR region\nstatic: each static logic IP is constrained to its relevant SLR\nall: enables both \'acc\' and \'static\' options\nBy default no floorplanning constraints are used', choices=['acc', 'static', 'all'], metavar='FLOORPLANNING_CONSTR')
        self.parser.add_argument('--placement_file', help='json file specifying accelerator placement', type=FileType())
        self.parser.add_argument('--slr_slices', help='enable SLR crossing register slices\nacc: create register slices for SLR crossing on accelerator-related interfaces\nstatic: create register slices for static logic IPs\nall: enable both \'acc\' and \'static\' options \nBy default they are disabled', choices=['acc', 'static', 'all'], metavar='SLR_SLICES')
        self.parser.add_argument('--interconnect_regslice', help='enable register slices on AXI interconnects\nall: enables them on all interconnects\nmem: enables them on interconnects in memory datapath\nhwruntime: enables them on the AXI-stream interconnects between the hwruntime and the accelerators\n', nargs='+', choices=['mem', 'hwruntime', 'all'], metavar='INTER_REGSLICE_LIST')
        self.parser.add_argument('--interconnect_opt', help='AXI interconnect optimization strategy: Minimize \'area\' or maximize \'performance\'\n(def: \'area\')', choices=['area', 'performance'], metavar='OPT_STRATEGY', action=StoreChoiceValue, default=0)
        self.parser.add_argument('--interconnect_priorities', help='enable priorities in the memory interconnect', action='store_true', default=False)
        self.parser.add_argument('--simplify_interconnection', help='simplify interconnection between accelerators and memory. Might negatively impact timing', action='store_true', default=False)
        self.parser.add_argument('--debug_intfs', help='choose which interfaces mark for debug and instantiate the correspondent ILA cores\nAXI: debug accelerator\'s AXI interfaces\nstream: debug accelerator\'s AXI-Stream interfaces\nboth: debug both accelerator\'s AXI and AXI-Stream interfaces\ncustom: debug user-defined interfaces\nnone: do not mark for debug any interface\n(def: \'none\')', choices=['AXI', 'stream', 'both', 'custom', 'none'], metavar='INTF_TYPE', default='none')
        self.parser.add_argument('--debug_intfs_list', help='path of file with the list of interfaces to debug', type=FileType())
        self.parser.add_argument('--ignore_eng_sample', help='ignore engineering sample status from chip part number', action='store_true', default=False)
        self.parser.add_argument('--target_language', help='choose target language to synthesize files to: VHDL or Verilog\n(def: \'VHDL\')', choices=['VHDL', 'Verilog'], metavar='TARGET_LANG', default='VHDL')

    def check_args(self, args):
        if args.debug_intfs == 'custom' and args.debug_intfs_list is None:
            msg.error('A file specifying which interfaces to mark for debug is required when choosing \'custom\' value on --debug_intfs argument')

        if args.interconnect_regslice is not None:
            for opt in args.interconnect_regslice:
                if opt == 'all' and len(args.interconnect_regslice) != 1:
                    msg.error('Invalid combination of values for --interconnect_regslice')

        if (args.slr_slices == 'acc' or args.slr_slices == 'all') and args.placement_file is None:
            msg.error('--placement_file argument required when enabling SLR-crossing register slices on accelerators')
        elif (args.floorplanning_constr == 'acc' or args.floorplanning_constr == 'all') and args.placement_file is None:
            msg.error('--placement_file argument required when setting floorplanning constraints on accelerators')

    def check_board_args(self, args, board):
        if args.memory_interleaving_stride is not None:
            if board.arch.device == 'zynq' or board.arch.device == 'zynqmp':
                msg.error('Memory interleaving is not available on neither Zynq nor ZynqMP boards')
            elif args.memory_interleaving_stride & (args.memory_interleaving_stride - 1):
                msg.error('Memory interleaving stride must be power of 2')
            elif math.log2(decimalFromHumanReadable(board.mem.bank_size)) - math.log2(args.memory_interleaving_stride) < math.ceil(math.log2(board.mem.num_banks)):
                msg.error('Max allowed interleaving stride in current board: ' + decimalToHumanReadable(2**(math.log2(decimalFromHumanReadable(board.mem.bank_size)) - math.ceil(math.log2(board.mem.num_banks))), 2))

        if args.enable_memory_bonding and board.mem.type != 'hbm':
            msg.error('Memory channel bonding is only available for HBM memories')

        if args.simplify_interconnection and (board.arch.device == 'zynq' or board.arch.device == 'zynqmp'):
            msg.error('Simplify memory interconnection is not available on neither Zynq nor ZynqMP boards')
        if args.simplify_interconnection and board.mem.type != 'ddr':
            msg.error('Simplify memory interconnection is only available for DDR memories')

        if not int(board.frequency.min) <= args.clock <= int(board.frequency.max):
            msg.error('Clock frequency requested (' + str(args.clock) + 'MHz) is not within the board range (' + str(board.frequency.min) + '-' + str(board.frequency.max) + 'MHz)')

        if (args.slr_slices is not None or args.floorplanning_constr is not None) and not hasattr(board.arch, 'slr'):
            msg.error('Use of placement constraints is only available for boards with SLRs')


parser = ArgParser()
help_text = parser.parser.format_help().split('\n')[0] + '\n' + '\n'.join(parser.parser.format_help().split('\n')[2:])



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import os
import shutil
import subprocess
import sys
import xml.etree.cElementTree as cET

import ait.backend.xilinx.utils.checkers as checkers
from ait.frontend.utils import ait_path, msg


script_folder = os.path.basename(os.path.dirname(os.path.realpath(__file__)))


def update_resource_utilization(acc):
    global available_resources
    global used_resources

    report_file = project_backend_path + '/HLS/' + acc.name + '/solution1/syn/report/' + acc.name + '_wrapper_csynth.xml'

    tree = cET.parse(report_file)
    root = tree.getroot()

    for resource in root.find('AreaEstimates').find('AvailableResources'):
        available_resources[resource.tag] = int(resource.text)

    if args.verbose_info:
        res_msg = 'Resources estimation for \'' + acc.name + '\': '
        res_msg += ', '.join(sorted(map(lambda r: r.tag + ' ' + r.text, list(root.find('AreaEstimates').find('Resources')))))
        msg.log(res_msg)

    depleted_resources = False
    error_message = 'Resource utilization over 100%\nResources estimation summary\n'
    for resource in root.find('AreaEstimates').find('Resources'):
        used_resources[resource.tag] = int(resource.text) * acc.num_instances + (int(used_resources[resource.tag]) if resource.tag in used_resources else 0)
        if used_resources[resource.tag] > available_resources[resource.tag]:
            if available_resources[resource.tag] == 0:
                msg.error('The HLS code is using resources not available in the selected FPGA')
            utilization_percentage = str(round(float(used_resources[resource.tag]) / float(available_resources[resource.tag]) * 100, 2))
            report_string = '{0:<9} {1:>7} used | {2:>7} available - {3:>6}% utilization\n'
            report_string_formatted = report_string.format(resource.tag, used_resources[resource.tag], available_resources[resource.tag], utilization_percentage)
            error_message += report_string_formatted
            depleted_resources = True

    if not args.disable_utilization_check and depleted_resources:
        msg.error(error_message.rstrip())


def synthesize_accelerator(acc):
    shutil.rmtree(project_backend_path + '/HLS/' + acc.name, ignore_errors=True)
    os.makedirs(project_backend_path + '/HLS/' + acc.name)

    shutil.copy2(acc.full_path, project_backend_path + '/HLS/' + acc.name + '/' + acc.filename)

    if (hls_executable == 'vitis_hls'):
        acc_tcl_script = '# Script automatically generated by the Accelerator Integration Tool. Edit at your own risk.\n' \
                         + 'open_project ' + acc.name + '\n' \
                         + 'set_top ' + acc.name + '_wrapper\n' \
                         + 'add_files ' + acc.name + '/' + acc.filename + ' -cflags "-I' + os.getcwd() + '"\n' \
                         + 'open_solution "solution1" -flow_target vivado\n' \
                         + 'set_part {' + chip_part + '}\n' \
                         + 'create_clock -period ' + str(args.clock) + 'MHz -name default\n' \
                         + 'config_rtl -reset_level low -reset_async\n' \
                         + 'config_interface -default_slave_interface none -m_axi_offset off'

        if board.arch.device == 'zynqmp' or board.arch.device == 'alveo':
            acc_tcl_script += ' -m_axi_addr64'

        acc_tcl_script += '\ncsynth_design\n' \
                          + 'export_design -rtl verilog -format ip_catalog -vendor bsc -library ompss -display_name ' + acc.name + ' -taxonomy /BSC/OmpSs\n' \
                          + 'exit\n'
    else:
        acc_tcl_script = '# Script automatically generated by the Accelerator Integration Tool. Edit at your own risk.\n' \
                         + 'open_project ' + acc.name + '\n' \
                         + 'set_top ' + acc.name + '_wrapper\n' \
                         + 'add_files ' + acc.name + '/' + acc.filename + ' -cflags "-I' + os.getcwd() + '"\n' \
                         + 'open_solution "solution1"\n' \
                         + 'set_part {' + chip_part + '} -tool vivado\n' \
                         + 'create_clock -period ' + str(args.clock) + 'MHz -name default\n' \
                         + 'config_rtl -reset control -reset_level low -reset_async\n'

        if board.arch.device == 'zynqmp' or board.arch.device == 'alveo':
            acc_tcl_script += 'config_interface -m_axi_addr64'

        acc_tcl_script += '\ncsynth_design\n' \
                          + 'export_design -rtl verilog -format ip_catalog -vendor bsc -library ompss -display_name ' + acc.name + ' -taxonomy /BSC/OmpSs\n' \
                          + 'exit\n'

    acc_tcl_script_file = open(project_backend_path + '/HLS/' + acc.name + '/HLS_' + acc.name + '.tcl', 'w')
    acc_tcl_script_file.write(acc_tcl_script)
    acc_tcl_script_file.close()

    msg.info('Synthesizing \'' + acc.name + '\'')

    p = subprocess.Popen(hls_executable + ' ' + project_backend_path + '/HLS/' + acc.name
                         + '/HLS_' + acc.name + '.tcl -l ' + project_backend_path
                         + '/HLS/' + acc.name + '/HLS_' + acc.name + '.log',
                         cwd=project_backend_path + '/HLS',
                         stdout=sys.stdout.subprocess, stderr=sys.stdout.subprocess, shell=True)
    if args.verbose:
        for line in iter(p.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))

    retval = p.wait()
    if retval:
        if not args.keep_files:
            shutil.rmtree(project_backend_path + '/HLS/' + acc.name, ignore_errors=True)
        msg.error('Synthesis of \'' + acc.name + '\' failed', start_time, False)
    else:
        msg.success('Finished synthesis of \'' + acc.name + '\'')

    update_resource_utilization(acc)


def run_step(project_args):
    global args
    global board
    global chip_part
    global start_time
    global ait_backend_path
    global project_backend_path
    global used_resources
    global available_resources
    global hls_executable

    args = project_args['args']
    board = project_args['board']
    start_time = project_args['start_time']
    project_path = project_args['path']
    accs = project_args['accs']

    chip_part = board.chip_part + ('-' + board.es if (board.es and not args.ignore_eng_sample) else '')
    ait_backend_path = ait_path + '/backend/' + args.backend
    project_backend_path = project_path + '/' + args.backend

    # Check if HLS requirements are met
    hls_executable = checkers.check_hls_tool()

    # Remove old directories used on the HLS step
    shutil.rmtree(project_backend_path + '/HLS', ignore_errors=True)
    os.makedirs(project_backend_path + '/HLS')

    msg.info('Synthesizing ' + str(args.num_accs) + ' accelerator' + ('s' if args.num_accs > 1 else ''))

    # Load used resources by hwruntime
    used_resources = dict()
    available_resources = dict()

    for acc in range(0, args.num_accs):
        synthesize_accelerator(accs[acc])

    if len(accs) > args.num_accs:
        msg.info('Synthesizing ' + str(len(accs) - args.num_accs) + ' additional auxiliary IP' + ('s' if len(accs) - args.num_accs > 1 else ''))

        for acc in range(args.num_accs, len(accs)):
            synthesize_accelerator(accs[acc])

    resources_file = open(project_path + '/' + args.name + '.resources-hls.txt', 'w')
    msg.log('Resources estimation summary')
    for res_name, res_value in sorted(used_resources.items()):
        if res_name in available_resources:
            available = available_resources[res_name]
        else:
            available = 0
        if available > 0:
            utilization_percentage = str(round(float(res_value) / float(available) * 100, 2))
            report_string = '{0:<9} {1:>7} used | {2:>7} available - {3:>6}% utilization'
            report_string_formatted = report_string.format(res_name, res_value, available, utilization_percentage)
            msg.log(report_string_formatted)
            resources_file.write(report_string_formatted + '\n')
    resources_file.close()



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import json
import os
import re
import shutil
import subprocess
import sys

import ait.backend.xilinx.utils.checkers as checkers
from ait.frontend.config import BITINFO_VERSION,  \
    VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH
from ait.frontend.utils import ait_path, decimalFromHumanReadable, msg

script_folder = os.path.basename(os.path.dirname(os.path.realpath(__file__)))


def generate_Vivado_variables_tcl():
    global accs
    global args

    vivado_project_variables = '# File automatically generated by the Accelerator Integration Tool. Edit at your own risk.\n' \
                               + '\n' \
                               + 'namespace eval AIT {' \
                               + '\n' \
                               + '\t# Project variables\n' \
                               + '\tvariable name_Project ' + args.name + '\n' \
                               + '\tvariable name_Design ' + args.name + '_design\n' \
                               + '\tvariable target_lang ' + args.target_language + '\n' \
                               + '\tvariable num_accs ' + str(args.num_instances) + '\n' \
                               + '\tvariable num_acc_creators ' + str(args.num_acc_creators) + '\n' \
                               + '\tvariable num_jobs ' + str(args.jobs) + '\n' \
                               + '\tvariable ait_call "' + str(re.escape(os.path.basename(sys.argv[0]) + ' ' + ' '.join(sys.argv[1:]))) + '"\n' \
                               + '\tvariable bitInfo_note ' + str(re.escape(args.bitinfo_note)) + '\n' \
                               + '\tvariable version_major_ait ' + str(VERSION_MAJOR) + '\n' \
                               + '\tvariable version_minor_ait ' + str(VERSION_MINOR) + '\n' \
                               + '\tvariable version_patch_ait ' + str(VERSION_PATCH) + '\n' \
                               + '\tvariable version_bitInfo ' + str(BITINFO_VERSION).lower() + '\n' \
                               + '\tvariable version_wrapper ' + (str(args.wrapper_version).lower() if args.wrapper_version else '0') + '\n' \
                               + '\tvariable memory_bonding ' + (str(args.enable_memory_bonding)) + '\n' \
                               + '\n' \
                               + '\t# IP caching variables\n' \
                               + '\tvariable IP_caching ' + str(not args.disable_IP_caching).lower() + '\n'

    if not args.disable_IP_caching:
        vivado_project_variables += '\tvariable path_CacheLocation ' + os.path.realpath(args.IP_cache_location) + '\n'

    regslice_all = '0'
    regslice_mem = '0'
    regslice_hwruntime = '0'
    if args.interconnect_regslice is not None:
        for opt in args.interconnect_regslice:
            if opt == 'all':
                regslice_all = '1'
            elif opt == 'mem':
                regslice_mem = '1'
            elif opt == 'hwruntime':
                regslice_hwruntime = '1'

    vivado_project_variables += '\n' \
                                + '\t# Bitstream variables\n' \
                                + '\tvariable interconOpt ' + str(args.interconnect_opt + 1) + '\n' \
                                + '\tvariable debugInterfaces ' + str(args.debug_intfs) + '\n' \
                                + '\tvariable interconRegSlice_all ' + regslice_all + '\n' \
                                + '\tvariable interconRegSlice_mem ' + regslice_mem + '\n' \
                                + '\tvariable interconRegSlice_hwruntime ' + regslice_hwruntime + '\n' \
                                + '\tvariable interleaving_stride ' + (hex(args.memory_interleaving_stride) if args.memory_interleaving_stride is not None else str(args.memory_interleaving_stride)) + '\n'\
                                + '\tvariable simplify_interconnection ' + str(args.simplify_interconnection).lower() + '\n' \
                                + '\tvariable interconPriority ' + str(args.interconnect_priorities) + '\n' \
                                + '\tvariable floorplanning_constr ' + str(args.floorplanning_constr) + '\n' \
                                + '\tvariable slr_slices ' + str(args.slr_slices) + '\n' \
                                + '\n' \
                                + '\t# ' + board.name + ' board variables\n' \
                                + '\tvariable board ' + board.name + '\n' \
                                + '\tvariable chipPart ' + chip_part + '\n'
    if board.board_part:
        vivado_project_variables += '\tvariable boardPart ' + str(board.board_part) + '\n'

    vivado_project_variables += '\tvariable clockFreq ' + str(args.clock) + '\n' \
                                + '\tvariable arch_device ' + board.arch.device + '\n'

    if args.slr_slices is not None or args.floorplanning_constr is not None:
        vivado_project_variables += '\tvariable board_slr_num ' + str(board.arch.slr.num) + '\n' \
                                    + '\tvariable board_slr_master ' + str(board.arch.slr.master) + '\n'

    vivado_project_variables += '\tvariable address_map [dict create]\n' \
                                + '\tdict set address_map "ompss_base_addr" ' + board.address_map.ompss_base_addr + '\n' \
                                + '\tdict set address_map "mem_base_addr" ' + board.address_map.mem_base_addr + '\n' \
                                + '\tdict set address_map "mem_type" ' + board.mem.type + '\n'

    if board.arch.device == 'zynq' or board.arch.device == 'zynqmp':
        vivado_project_variables += '\tdict set address_map "mem_size" ' + hex(decimalFromHumanReadable(board.mem.size)) + '\n'
    elif board.arch.device == 'alveo':
        vivado_project_variables += '\tdict set address_map "mem_num_banks" ' + str(board.mem.num_banks) + '\n' \
                                    + '\tdict set address_map "mem_bank_size" ' + hex(decimalFromHumanReadable(board.mem.bank_size)) + '\n'

    vivado_project_variables += '\n' \
                                + '\t# Hardware Instrumentation variables\n' \
                                + '\tvariable hwcounter ' + str(args.hwcounter) + '\n' \
                                + '\tvariable hwinst ' + str(args.hwinst) + '\n'

    vivado_project_variables += '\n' \
                                + '\t# HW runtime variables\n' \
                                + '\tvariable deps_hwruntime ' + str(args.deps_hwruntime) + '\n' \
                                + '\tvariable task_creation ' + str(args.task_creation) + '\n' \
                                + '\tvariable lock_hwruntime ' + str(args.lock_hwruntime) + '\n' \
                                + '\tvariable cmdInSubqueue_len ' + str(args.cmdin_subqueue_len) + '\n' \
                                + '\tvariable cmdOutSubqueue_len ' + str(args.cmdout_subqueue_len) + '\n' \
                                + '\tvariable spawnInQueue_len ' + str(args.spawnin_queue_len) + '\n' \
                                + '\tvariable spawnOutQueue_len ' + str(args.spawnout_queue_len) + '\n' \
                                + '\tvariable hwruntime_interconnect ' + str(args.hwruntime_interconnect) + '\n' \
                                + '\tvariable enable_spawn_queues ' + str(not args.disable_spawn_queues) + '\n' \
                                + '\tvariable max_args_per_task ' + str(args.max_args_per_task) + '\n' \
                                + '\tvariable max_deps_per_task ' + str(args.max_deps_per_task) + '\n' \
                                + '\tvariable max_copies_per_task ' + str(args.max_copies_per_task) + '\n'
    if args.deps_hwruntime:
        vivado_project_variables += '\n' \
                                    + '\t# Picos parameters\n' \
                                    + '\tvariable picos_num_dcts ' + str(args.picos_num_dcts) + '\n' \
                                    + '\tvariable picos_tm_size ' + str(args.picos_tm_size) + '\n' \
                                    + '\tvariable picos_dm_size ' + str(args.picos_dm_size) + '\n' \
                                    + '\tvariable picos_vm_size ' + str(args.picos_vm_size) + '\n' \
                                    + '\tvariable picos_dm_ds ' + args.picos_dm_ds + '\n' \
                                    + '\tvariable picos_dm_hash ' + args.picos_dm_hash + '\n' \
                                    + '\tvariable picos_hash_t_size ' + str(args.picos_hash_t_size) + '\n'

    vivado_project_variables += '\n' \
                                + '\t# List of accelerators\n' \
                                + '\tset accs [list'

    for acc in accs[0:args.num_accs]:
        acc_name = str(acc.type) + ':' + str(acc.num_instances) + ':' + acc.name + ':' + str(acc.task_creation)
        vivado_project_variables += ' ' + acc_name

    vivado_project_variables += '\t]\n'

    # Generate acc instance list with SLR info
    # Placement info is only needed for registers, constraints are dumped into constraint file
    if (args.slr_slices == 'acc') or (args.slr_slices == 'all'):
        acc_pl_dict = 'set acc_placement [dict create '
        for acc in accs[0:args.num_accs]:
            if hasattr(acc, 'SLR'):
                acc_pl_dict += ' ' + str(acc.name) + ' [list'
                for slrnum in acc.SLR:
                    acc_pl_dict += ' ' + str(slrnum)
                acc_pl_dict += ']'
        acc_pl_dict += ']'
        vivado_project_variables += '\t' + acc_pl_dict + '\n'

    # Generate acc constraint file
    if (args.floorplanning_constr == 'acc') or (args.floorplanning_constr == 'all'):
        accConstrFiles = open(f'{project_backend_path}/board/{board.name}/constraints/acc_floorplan.xdc', 'w')
        for acc in accs[0:args.num_accs]:
            if hasattr(acc, 'SLR'):
                instancesToPlace = len(acc.SLR)
                if len(acc.SLR) > acc.num_instances:
                    instancesToPlace = acc.num_instances
                    msg.warning('Placement list for accelerator {} has more instances than expected ({} > {}). Placing instances 0-{}'.format(acc.name, len(acc.SLR), acc.num_instances, instancesToPlace - 1))
                elif len(acc.SLR) < acc.num_instances:
                    instancesToPlace = len(acc.SLR)
                    msg.warning('Placement list for accelerator {} has less instances than expected ({} < {}). Placing instances 0-{}'.format(acc.name, len(acc.SLR), acc.num_instances, instancesToPlace - 1))
                # Instantiate each accelerator with a single instance and placement info
                for instanceNumber in range(instancesToPlace):
                    accBlock = f'{acc.name}_{instanceNumber}'
                    accConstrFiles.write(f'add_cells_to_pblock [get_pblocks slr{acc.SLR[instanceNumber]}_pblock] '
                                         + '[get_cells {'
                                         + f'*/{accBlock}/Adapter_* '
                                         + f'*/{accBlock}/*_addrInterleaver '
                                         + f'*/{accBlock}/TID_subset_converter '
                                         + f'*/{accBlock}/{acc.name}_ompss'
                                         + '}]\n')
                    if acc.task_creation:
                        accConstrFiles.write(f'add_cells_to_pblock [get_pblocks slr{acc.SLR[instanceNumber]}_pblock] '
                                             + '[get_cells {'
                                             + f'*/{accBlock}/new_task_spawner '
                                             + f'*/{accBlock}/axis_tid_demux '
                                             + '}]\n')
        accConstrFiles.close()

    if args.datainterfaces_map and os.path.exists(args.datainterfaces_map):
        if args.verbose_info:
            msg.log('Parsing user data interfaces map: ' + args.datainterfaces_map)

        vivado_project_variables += '\n' \
                                    + '\t# List of datainterfaces map\n' \
                                    + '\tSet dataInterfaces_map [list'

        with open(args.datainterfaces_map) as map_file:
            map_data = map_file.readlines()
            for map_line in map_data:
                elems = map_line.strip().replace('\n', '').split('\t')
                if len(elems) >= 2 and len(elems[0]) > 0 and elems[0][0] != '#':
                    vivado_project_variables += '\t {' + elems[0] + ' ' + elems[1] + '}'

        vivado_project_variables += '\t]\n'
    elif args.datainterfaces_map:
        msg.error('User data interfaces map not found: ' + args.datainterfaces_map)
    else:
        vivado_project_variables += '\n' \
                                    + '\t# List of datainterfaces map\n' \
                                    + '\tset dataInterfaces_map [list]\n'

    if args.debug_intfs == 'custom' and os.path.exists(args.debug_intfs_list):
        if args.verbose_info:
            msg.log('Parsing user-defined interfaces to debug: ' + args.debug_intfs_list)

        vivado_project_variables += '\n' \
                                    + '\t# List of debugInterfaces list\n' \
                                    + '\tset debugInterfaces_list [list'

        with open(args.debug_intfs_list) as map_file:
            map_data = map_file.readlines()
            for map_line in map_data:
                elems = map_line.strip().replace('\n', '')
                if elems[0][0] != '#':
                    vivado_project_variables += '\t ' + str(elems)

        vivado_project_variables += ']\n'
    elif args.debug_intfs == 'custom':
        msg.error('User-defined interfaces to debug file not found: ' + args.debug_intfs_list)

    vivado_project_variables += '}\n'

    vivado_project_variables_file = open(project_backend_path + '/tcl/projectVariables.tcl', 'w')
    vivado_project_variables_file.write(vivado_project_variables)
    vivado_project_variables_file.close()


def load_acc_placement(accList, args):
    # Read placement info from file
    if args.placement_file and os.path.exists(args.placement_file):
        usrPlacement = json.load(open(args.placement_file))
        for acc in accList:
            if acc.name not in usrPlacement:
                msg.warning('No placement given for acc ' + acc.name)
            else:
                acc.SLR = usrPlacement[acc.name]

    elif args.placement_file:
        msg.error('Placement file not found: ' + args.user_constraints)


def run_step(project_args):
    global args
    global board
    global accs
    global chip_part
    global start_time
    global ait_backend_path
    global project_backend_path

    args = project_args['args']
    board = project_args['board']
    accs = project_args['accs']
    start_time = project_args['start_time']
    project_path = project_args['path']

    chip_part = board.chip_part + ('-' + board.es if (board.es and not args.ignore_eng_sample) else '')
    ait_backend_path = ait_path + '/backend/' + args.backend
    project_backend_path = project_path + '/' + args.backend

    # Check if Vivado requirements are met
    checkers.check_vivado()

    # Load accelerator placement info
    load_acc_placement(accs[0:args.num_accs], args)

    if args.memory_interleaving_stride is not None:
        subprocess.check_output(['sed -i "s/\`undef __ENABLE__/\`define __ENABLE__/" ' + project_backend_path + '/IPs/addrInterleaver.v'], shell=True)

    if args.user_constraints and os.path.exists(args.user_constraints):
        constraints_path = project_backend_path + '/board/' + board.name + '/constraints'
        if not os.path.exists(constraints_path):
            os.mkdir(constraints_path)
        if args.verbose_info:
            msg.log('Adding user constraints file: ' + args.user_constraints)
        shutil.copy2(args.user_constraints, constraints_path + '/')
    elif args.user_constraints:
        msg.error('User constraints file not found: ' + args.user_constraints)

    if args.user_pre_design and os.path.exists(args.user_pre_design):
        user_pre_design_ext = args.user_pre_design.split('.')[-1] if len(args.user_pre_design.split('.')) > 1 else ''
        if user_pre_design_ext != 'tcl':
            msg.error('Invalid extension for PRE design TCL script: ' + args.user_pre_design)
        elif args.verbose_info:
            msg.log('Adding pre design user script: ' + args.user_pre_design)
        shutil.copy2(args.user_pre_design, project_backend_path + '/tcl/scripts/userPreDesign.tcl')
    elif args.user_pre_design:
        msg.error('User PRE design TCL script not found: ' + args.user_pre_design)

    if args.user_post_design and os.path.exists(args.user_post_design):
        user_post_design_ext = args.user_post_design.split('.')[-1] if len(args.user_post_design.split('.')) > 1 else ''
        if user_post_design_ext != 'tcl':
            msg.error('Invalid extension for POST design TCL script: ' + args.user_post_design)
        elif args.verbose_info:
            msg.log('Adding post design user script: ' + args.user_post_design)
        shutil.copy2(args.user_post_design, project_backend_path + '/tcl/scripts/userPostDesign.tcl')
    elif args.user_post_design:
        msg.error('User POST design TCL script not found: ' + args.user_post_design)

    # Generate tcl file with project variables
    generate_Vivado_variables_tcl()

    # Enable beta device on Vivado init script
    if os.path.exists(project_backend_path + '/board/' + board.name + '/board_files'):
        p = subprocess.Popen('echo "enable_beta_device ' + chip_part + '\nset_param board.repoPaths [list '
                             + project_backend_path + '/board/' + board.name + '/board_files]" > '
                             + project_backend_path + '/vivado.tcl', shell=True)
        retval = p.wait()
    else:
        p = subprocess.Popen('echo "enable_beta_device ' + chip_part + '" > '
                             + project_backend_path + '/vivado.tcl', shell=True)
        retval = p.wait()

    p = subprocess.Popen('vivado -init -nojournal -nolog -notrace -mode batch -source '
                         + project_backend_path + '/tcl/scripts/generate_design.tcl',
                         cwd=project_backend_path, stdout=sys.stdout.subprocess,
                         stderr=sys.stdout.subprocess, shell=True)

    if args.verbose:
        for line in iter(p.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))

    retval = p.wait()
    if retval:
        msg.error('Block Design generation failed', start_time, False)
    else:
        msg.success('Block Design generated')



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import os
import unittest
from unittest.mock import ANY, MagicMock, patch

from ait.frontend.parser import ArgParser


class Test_IP_caching(unittest.TestCase):
    def test_disabled_exists(self):
        # Create the args object which could contain any attribute
        args = type('', (), {})()
        args.from_step = 'HLS'
        args.to_step = 'HLS'
        args.backend = 'xilinx'
        args.disable_IP_caching = True
        args.IP_cache_location = '/tmp'

        parser = ArgParser()
        parser.check_flow_args(args)

        self.assertTrue(True)  # Just check no errors are thrown

    def test_disabled_not_exists(self):
        # Create the args object which could contain any attribute
        args = type('', (), {})()
        args.from_step = 'HLS'
        args.to_step = 'HLS'
        args.backend = 'xilinx'
        args.disable_IP_caching = True
        args.IP_cache_location = '/this-path-should-not-exist'

        parser = ArgParser()
        parser.check_flow_args(args)

        self.assertTrue(True)  # Just check no errors are thrown

    def test_enabled_exists(self):
        # Create the args object which could contain any attribute
        args = type('', (), {})()
        args.from_step = 'HLS'
        args.to_step = 'HLS'
        args.backend = 'xilinx'
        args.disable_IP_caching = False
        args.IP_cache_location = '/tmp'

        parser = ArgParser()
        parser.check_flow_args(args)

        self.assertTrue(True)  # Just check no errors are thrown

    @patch('ait.frontend.utils.msg.error')
    def test_enabled_not_exists(self, msg_error):
        # Create the args object which could contain any attribute
        args = type('', (), {})()
        args.from_step = 'HLS'
        args.to_step = 'HLS'
        args.backend = 'xilinx'
        args.disable_IP_caching = False
        args.IP_cache_location = '/this-path-should-not-exist'

        parser = ArgParser()
        parser.is_default = MagicMock(return_value=False)
        parser.check_flow_args(args)

        parser.is_default.assert_called_with('IP_cache_location', args.backend)
        msg_error.assert_called_with(ANY)

    def test_enabled_default(self):
        # Create the args object which could contain any attribute
        args = type('', (), {})()
        args.from_step = 'HLS'
        args.to_step = 'HLS'
        args.backend = 'xilinx'
        args.disable_IP_caching = False
        args.IP_cache_location = '/tmp/ait-test-folder'

        # Check PRE conditions
        if (os.path.exists(args.IP_cache_location)):
            os.removedirs(args.IP_cache_location)
        self.assertFalse(os.path.exists(args.IP_cache_location))

        parser = ArgParser()
        parser.is_default = MagicMock(return_value=True)
        parser.check_flow_args(args)

        # Check POST conditions
        parser.is_default.assert_called_with('IP_cache_location', args.backend)
        self.assertTrue(os.path.exists(args.IP_cache_location))
        os.rmdir(args.IP_cache_location)

    def test_enabled_default_exists(self):
        # Create the args object which could contain any attribute
        args = type('', (), {})()
        args.from_step = 'HLS'
        args.to_step = 'HLS'
        args.backend = 'xilinx'
        args.disable_IP_caching = False
        args.IP_cache_location = '/tmp/ait-test-folder'

        # Check PRE conditions
        if (not os.path.exists(args.IP_cache_location)):
            os.makedirs(args.IP_cache_location)

        parser = ArgParser()
        parser.is_default = MagicMock(return_value=True)
        parser.check_flow_args(args)

        # Check POST conditions
        self.assertTrue(os.path.exists(args.IP_cache_location))
        os.rmdir(args.IP_cache_location)


if __name__ == '__main__':
    unittest.main()



import os

NAME = 'xilinx'

MIN_VITIS_HLS_VERSION = '2020.2'
MIN_VIVADO_HLS_VERSION = '2018.3'
MIN_VIVADO_VERSION = '2018.3'


def get_supported_boards():
    return sorted(next(os.walk(os.path.dirname(__file__) + '/board'))[1])


info = dict()
info['boards'] = get_supported_boards()
info['steps'] = ['HLS', 'design', 'synthesis', 'implementation', 'bitstream', 'boot']
info['initial_step'] = 'HLS'
info['final_step'] = 'bitstream'



import importlib
import json
import os
import shutil

import ait.backend.xilinx.utils.checkers as checkers
import ait.backend.xilinx.utils.parser as parser
from ait.frontend.utils import ait_path, msg


class JSONObject:
    def __init__(self, dict):
        vars(self).update(dict)


def load(args):

    project_path = os.path.normpath(os.path.realpath(args.dir + '/' + args.name + '_ait'))
    project_backend_path = project_path + '/' + args.backend
    ait_backend_path = ait_path + '/backend/' + args.backend

    board = json.load(open(ait_path + '/backend/' + args.backend + '/board/' + args.board + '/basic_info.json'), object_hook=JSONObject)

    # Check backend-related board arguments
    parser.parser.check_board_args(args, board)

    # Check for backend support for the given board
    if not args.disable_board_support_check:
        if args.verbose_info:
            msg.log('Checking vendor support for selected board')

        chip_part = board.chip_part + ('-' + board.es if board.es and not args.ignore_eng_sample else '')

        checkers.check_board_support(chip_part)

    shutil.rmtree(project_backend_path + '/tcl', ignore_errors=True)
    shutil.copytree(ait_backend_path + '/tcl', project_backend_path + '/tcl')

    shutil.rmtree(project_backend_path + '/IPs', ignore_errors=True)
    shutil.copytree(ait_backend_path + '/IPs', project_backend_path + '/IPs')

    shutil.rmtree(project_backend_path + '/board', ignore_errors=True)
    shutil.copytree(ait_backend_path + '/board/' + args.board, project_backend_path + '/board/' + args.board)

    steps = importlib.import_module('ait.backend.xilinx.steps')

    return steps, board



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import os
import shutil
import subprocess

from ait.backend.xilinx.info import MIN_VITIS_HLS_VERSION, MIN_VIVADO_HLS_VERSION, MIN_VIVADO_VERSION
from ait.frontend.utils import msg

vivado_version = None
hls_version = None


def check_vivado():
    global vivado_version

    if shutil.which('vivado'):
        vivado_version = str(subprocess.check_output(['vivado -version | head -n1 | sed "s/\(Vivado.\+v\)\(\([0-9]\|\.\)\+\).\+/\\2/"'], shell=True), 'utf-8').strip()
        if vivado_version < MIN_VIVADO_VERSION:
            msg.error('Installed Vivado version ({}) not supported (>= {})'.format(vivado_version, MIN_VIVADO_VERSION))
    else:
        msg.error('vivado not found. Please set PATH correctly')

    return True


def check_hls_tool():
    if shutil.which('vivado_hls'):
        check_vivado_hls()
        return 'vivado_hls'
    elif shutil.which('vitis_hls'):
        check_vitis_hls()
        return 'vitis_hls'
    else:
        msg.error('No HLS tool found. Please set PATH correctly')


def check_vivado_hls():
    global hls_version

    if shutil.which('vivado_hls'):
        hls_version = str(subprocess.check_output(['vivado_hls -version | head -n1 | sed "s/\(Vivado.\+v\)\(\([0-9]\|\.\)\+\).\+/\\2/"'], shell=True), 'utf-8').strip()
        if hls_version < MIN_VIVADO_HLS_VERSION:
            msg.error('Installed Vivado HLS version ({}) not supported (>= {})'.format(hls_version, MIN_VIVADO_HLS_VERSION))
    else:
        msg.warning('vivado_hls not found. Please set PATH correctly')

    return True


def check_vitis_hls():
    global hls_version

    if shutil.which('vitis_hls'):
        hls_version = str(subprocess.check_output(['vitis_hls -version | head -n1 | sed "s/\(Vitis.\+v\)\(\([0-9]\|\.\)\+\).\+/\\2/"'], shell=True), 'utf-8').strip()
        if hls_version < MIN_VITIS_HLS_VERSION:
            msg.error('Installed Vitis HLS version ({}) not supported (>= {})'.format(hls_version, MIN_VITIS_HLS_VERSION))
    else:
        msg.error('vitis_hls not found. Please set PATH correctly')

    return True


def check_bootgen():
    if shutil.which('bootgen'):
        msg.warning('bootgen not found. .bit.bin file will not be generated')
        return False

    return True


def check_petalinux(petalinux_build_path, petalinux_install_path):
    if (not os.path.exists(petalinux_build_path) or not os.path.exists(petalinux_install_path)):
        msg.error('PETALINUX_BUILD (' + (petalinux_build_path if petalinux_build_path else 'empty') + ') or PETALINUX_INSTALL ('
                  + (petalinux_install_path if petalinux_install_path else 'empty') + ') variables not properly set')
        msg.error('Generation of petalinux boot files failed')

    env = str(subprocess.Popen('bash -c "trap \'env\' exit; source ' + petalinux_install_path
                               + '/settings.sh > /dev/null 2>&1"', shell=True,
                               stdout=subprocess.PIPE).communicate()[0], 'utf-8').strip('\n')

    # NOTE: Only importing some environment variables as there may be complex functions/expansions that
    #       we do not need to handle here
    for line in env.split('\n'):
        splitted = line.split('=', 1)
        if splitted[0] == 'PATH' or splitted[0].find('PETALINUX') != -1:
            os.environ.update(dict([line.split('=', 1)]))

    if not shutil.which('petalinux-config'):
        msg.error('petalinux commands not found. Please check PETALINUX_INSTALL environment variable')

    return True


def check_board_support(chip_part):
    check_vivado()

    tmp_dir = os.popen('mktemp -d --suffix=_ait').read().rstrip()

    os.mkdir(tmp_dir + '/scripts')

    os.system('echo "enable_beta_device ' + chip_part + '" >' + tmp_dir + '/scripts/vivado.tcl')
    os.system('echo "if {[llength [get_parts ' + chip_part + ']] == 0} {exit 1}" > ' + tmp_dir + '/scripts/ait_part_check.tcl')
    p = subprocess.Popen('vivado -init -nojournal -nolog -mode batch -source ' + tmp_dir + '/scripts/ait_part_check.tcl', shell=True, stdout=open(os.devnull, 'w'), cwd=tmp_dir + '/scripts')
    retval = p.wait()
    os.system('rm -rf ' + tmp_dir)
    if (int(retval) == 1):
        msg.error('Your current version of Vivado does not support part ' + chip_part)



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import os
import subprocess
import sys

import ait.backend.xilinx.utils.checkers as checkers
from ait.frontend.utils import ait_path, msg

script_folder = os.path.basename(os.path.dirname(os.path.realpath(__file__)))


def run_step(project_args):
    global args
    global board
    global chip_part
    global start_time
    global ait_backend_path
    global project_backend_path

    args = project_args['args']
    board = project_args['board']
    start_time = project_args['start_time']
    project_path = project_args['path']

    chip_part = board.chip_part + ('-' + board.es if (board.es and not args.ignore_eng_sample) else '')
    ait_backend_path = ait_path + '/backend/' + args.backend
    project_backend_path = project_path + '/' + args.backend

    # Check if the requirements are met
    checkers.check_vivado()

    if os.path.isfile(project_backend_path + '/' + args.name + '/' + args.name + '.xpr'):
        # Enable beta device on Vivado init script
        if board.board_part:
            p = subprocess.Popen('echo "enable_beta_device ' + chip_part + '\nset_param board.repoPaths [list '
                                 + project_backend_path + '/board/' + board.name + '/board_files]" > '
                                 + project_backend_path + '/vivado.tcl', shell=True)
            retval = p.wait()
        else:
            p = subprocess.Popen('echo "enable_beta_device ' + chip_part + '" > '
                                 + project_backend_path + '/vivado.tcl', shell=True)
            retval = p.wait()

        p = subprocess.Popen('vivado -init -nojournal -nolog -notrace -mode batch -source '
                             + project_backend_path + '/tcl/scripts/implement_design.tcl',
                             cwd=project_backend_path,
                             stdout=sys.stdout.subprocess,
                             stderr=sys.stdout.subprocess, shell=True)

        if args.verbose:
            for line in iter(p.stdout.readline, b''):
                sys.stdout.write(line.decode('utf-8'))

        retval = p.wait()
        if retval:
            msg.error('Hardware implementation failed', start_time, False)
        else:
            msg.success('Hardware implemented')
    else:
        msg.error('No Vivado .xpr file exists for the current project. Hardware implementation failed')



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import argparse
import importlib
import json
import math
import os
import re
import subprocess
import sys

from ait.frontend.config import LONG_VERSION, MIN_WRAPPER_VERSION
from ait.frontend.utils import backends, decimalFromHumanReadable, \
    decimalToHumanReadable, msg

for backend in backends:
    importlib.import_module('ait.backend.%s.info' % (backend))


# Custom argparse type representing a power of 2 int
class IntPowerType:
    def __init__(self, imin=None, imax=None):
        self.imin = imin
        self.imax = imax

    def __call__(self, arg):
        try:
            value = int(arg)
        except ValueError:
            raise self.exception()
        if (self.imin is not None and value < self.imin) or (self.imax is not None and value > self.imax):
            raise self.exception()
        elif value & (value - 1):
            raise self.exception(value)
        return value

    def exception(self, value=None):
        if value is not None and (value & (value - 1)):
            return argparse.ArgumentTypeError('must be power of 2')
        elif self.imin is not None and self.imax is not None:
            return argparse.ArgumentTypeError('must be an integer power of 2 in the range [{}, {}]'.format(self.imin, self.imax))
        elif self.imin is not None:
            return argparse.ArgumentTypeError('must be an integer power of 2 >= {}'.format(self.imin))
        elif self.imax is not None:
            return argparse.ArgumentTypeError('must be an integer power of 2 <= {}'.format(self.imax))
        else:
            return argparse.ArgumentTypeError('must be an integer power of 2')


# Custom argparse type representing a bounded int
class IntRangeType:
    def __init__(self, imin=None, imax=None):
        self.imin = imin
        self.imax = imax

    def __call__(self, arg):
        try:
            value = int(arg)
        except ValueError:
            raise self.exception()
        if (self.imin is not None and value < self.imin) or (self.imax is not None and value > self.imax):
            raise self.exception()
        return value

    def exception(self):
        if self.imin is not None and self.imax is not None:
            return argparse.ArgumentTypeError('must be an integer in the range [{}, {}]'.format(self.imin, self.imax))
        elif self.imin is not None:
            return argparse.ArgumentTypeError('must be an integer >= {}'.format(self.imin))
        elif self.imax is not None:
            return argparse.ArgumentTypeError('must be an integer <= {}'.format(self.imax))
        else:
            return argparse.ArgumentTypeError('must be an integer')


# Custom argparse type representing a human readable number
class HumanReadableType:
    def __init__(self, vmin=None, vmax=None):
        self.vmin = vmin
        self.vmax = vmax
        if vmin is not None:
            self.vmin = decimalFromHumanReadable(vmin)
        if vmax is not None:
            self.vmax = decimalFromHumanReadable(vmax)

    def __call__(self, arg):
        try:
            value = decimalFromHumanReadable(arg)
        except ValueError:
            raise self.valueException()
        except TypeError:
            raise self.typeException()
        if (self.vmin is not None and value < self.vmin) or (self.vmax is not None and value > self.vmax):
            raise self.rangeException()
        return value

    def typeException(self):
        return argparse.ArgumentTypeError('invalid unit')

    def valueException(self):
        return argparse.ArgumentTypeError('invalid value')

    def rangeException(self):
        if self.vmin is not None and self.vmax is not None:
            return argparse.ArgumentTypeError('value must be in the range [{}, {}]'.format(decimalToHumanReadable(self.vmin), decimalToHumanReadable(self.vmax)))
        elif self.vmin is not None:
            return argparse.ArgumentTypeError('value must be >= {}'.format(decimalToHumanReadable(self.vmin)))
        elif self.vmax is not None:
            return argparse.ArgumentTypeError('value must be <= {}'.format(decimalToHumanReadable(self.vmax)))


# Custom argparse type representing a path
class PathType:
    def __call__(self, arg):
        if os.path.isdir(arg):
            return os.path.realpath(arg)
        else:
            raise argparse.ArgumentTypeError('invalid path')


# Custom argparse type representing a path to a file
class FileType:
    def __call__(self, arg):
        if os.path.isfile(arg):
            return os.path.realpath(arg)
        else:
            raise argparse.ArgumentTypeError('invalid file')


class StoreChoiceValue(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.choices.index(values))


class ChangedArg(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        msg.warning('Argument ' + '/'.join(self.option_strings) + ' has changed. Check `ait -h` and fix your program call')


class RemovedArg(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        msg.info('Argument ' + '/'.join(self.option_strings) + ' is no longer used. Please remove it from the program call')


class CustomParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('usage: ait -b BOARD -n NAME\nait error: %s\n' % message)
        sys.exit(2)


def getNumJobs():
    # NOTE: assuming at most 3GB of memory usage per job
    procsByMem = int(subprocess.check_output(["free -b | grep 'Mem:' | awk {'print int(($7/1024^3)/3)'}"], shell=True))
    nprocs = int(subprocess.check_output(['nproc']))

    return max(1, min(procsByMem, nprocs))


class ArgParser:
    defaults = dict()
    for backend in backends:
        defaults[backend] = dict()

        # IP cache location defaults
        defaults[backend]['IP_cache_location'] = {
            'value': '/var/tmp/ait/' + backend + '/IP_cache',
            'used': False
        }

        defaults[backend]['hwinst'] = {
            'value': False,
            'used': False
        }

    def __init__(self):
        # Load vendor-specific parsers
        self.backend_parser = dict()
        for backend in backends:
            backend_parser = 'ait.backend.{}.utils.parser'.format(backend)
            if importlib.util.find_spec(backend_parser):
                self.backend_parser[backend] = importlib.import_module(backend_parser)

        # Create main parser
        self.parser = CustomParser(formatter_class=argparse.RawTextHelpFormatter, epilog='\n'.join(self.backend_parser[backend].help_text for backend in self.backend_parser), prog='ait', add_help=False, usage='%(prog)s -b BOARD -n NAME\nThe Accelerator Integration Tool (AIT) automatically integrates OmpSs@FPGA accelerators into FPGA designs using different vendor backends.')

        # Required arguments
        required_args = self.parser.add_argument_group('Required')
        required_args.add_argument('-b', '--board', help='board model. Supported boards by vendor:\n' + '\n'.join(backend + ': ' + ', '.join(board for board in backends[backend]['boards']) for backend in backends), choices=[board for board in backends[backend]['boards'] for backend in backends], metavar='BOARD', type=str.lower, required=True)
        required_args.add_argument('-n', '--name', help='project name', metavar='NAME', required=True)

        # Generation flow arguments
        flow_args = self.parser.add_argument_group('Generation flow')
        flow_args.add_argument('-d', '--dir', help='path where the project directory tree will be created\n(def: \'./\')', type=PathType(), default='./')
        flow_args.add_argument('--disable_IP_caching', help='disable IP caching. Significantly increases generation time', action='store_true', default=False)
        flow_args.add_argument('--disable_utilization_check', help='disable resources utilization check during HLS generation', action='store_true', default=False)
        flow_args.add_argument('--disable_board_support_check', help='disable board support check', action='store_true', default=False)
        flow_args.add_argument('--from_step', help='initial generation step. Generation steps by vendor:\n' + '\n'.join(backend + ': ' + ', '.join(step for step in backends[backend]['steps']) + "\n(def: '{}')".format(backends[backend]['initial_step']) for backend in backends), choices=[step for step in backends[backend]['steps'] for backend in backends], metavar='FROM_STEP', default='HLS')
        flow_args.add_argument('--IP_cache_location', help='path where the IP cache will be located\n(def: \'/var/tmp/ait/<vendor>/IP_cache/\')', type=PathType())
        flow_args.add_argument('--to_step', help='final generation step. Generation steps by vendor:\n' + '\n'.join(backend + ': ' + ', '.join(step for step in backends[backend]['steps']) + "\n(def: '{}')".format(backends[backend]['final_step']) for backend in backends), choices=[step for step in backends[backend]['steps'] for backend in backends], metavar='TO_STEP', default='bitstream')

        # Bitstream configuration arguments
        bitstream_args = self.parser.add_argument_group('Bitstream configuration')
        bitstream_args.add_argument('-c', '--clock', help='FPGA clock frequency in MHz\n(def: \'100\')', type=int, default='100')
        bitstream_args.add_argument('--hwruntime', help=argparse.SUPPRESS, action=RemovedArg)
        bitstream_args.add_argument('--hwcounter', help='add a hardware counter to the bitstream', action='store_true', default=False)
        bitstream_args.add_argument('--wrapper_version', help='version of accelerator wrapper shell. This information will be placed in the bitstream information', type=int)
        bitstream_args.add_argument('--bitinfo_note', help='custom note to add to the bitInfo', type=ascii, default='')

        # Data path arguments
        datapath_args = self.parser.add_argument_group('Data path')
        datapath_args.add_argument('--datainterfaces_map', help='path of mappings file for the data interfaces', type=FileType())
        datapath_args.add_argument('--memory_interleaving_stride', help='size in bytes of the stride of the memory interleaving. By default there is no interleaving', metavar='MEM_INTERLEAVING_STRIDE', type=HumanReadableType(vmin='4K'))
        datapath_args.add_argument('--enable_memory_bonding', help='bond memory channels to increase data throughput. By default there is no bonding', action='store_true', default=False)

        # Hardware Runtime arguments
        hwruntime_args = self.parser.add_argument_group('Hardware Runtime')
        hwruntime_args.add_argument('--cmdin_queue_len', help='maximum length (64-bit words) of the queue for the hwruntime command in\nThis argument is mutually exclusive with --cmdin_subqueue_len', type=IntRangeType(imin=4))
        hwruntime_args.add_argument('--cmdin_subqueue_len', help='length (64-bit words) of each accelerator subqueue for the hwruntime command in.\nThis argument is mutually exclusive with --cmdin_queue_len\nMust be power of 2\nDef. max(64, 1024/num_accs)', type=IntPowerType(imin=4))
        hwruntime_args.add_argument('--cmdout_queue_len', help='maximum length (64-bit words) of the queue for the hwruntime command out\nThis argument is mutually exclusive with --cmdout_subqueue_len', type=IntRangeType(imin=2))
        hwruntime_args.add_argument('--cmdout_subqueue_len', help='length (64-bit words) of each accelerator subqueue for the hwruntime command out. This argument is mutually exclusive with --cmdout_queue_len\nMust be power of 2\nDef. max(64, 1024/num_accs)', type=IntPowerType(imin=2))
        hwruntime_args.add_argument('--disable_spawn_queues', help='disable the hwruntime spawn in/out queues', action='store_true', default=False)
        hwruntime_args.add_argument('--spawnin_queue_len', help='length (64-bit words) of the hwruntime spawn in queue. Must be power of 2\n(def: \'1024\')', type=IntPowerType(imin=4), default=1024)
        hwruntime_args.add_argument('--spawnout_queue_len', help='length (64-bit words) of the hwruntime spawn out queue. Must be power of 2\n(def: \'1024\')', type=IntPowerType(imin=4), default=1024)
        hwruntime_args.add_argument('--hwruntime_interconnect', help='type of hardware runtime interconnection with accelerators\ncentralized\ndistributed\n(def: \'centralized\')', choices=['centralized', 'distributed'], metavar='HWR_INTERCONNECT', default='centralized')  # TODO: Explain what does each option do
        hwruntime_args.add_argument('--max_args_per_task', help='maximum number of arguments for any task in the bitstream\n(def: \'15\')', type=IntRangeType(imin=1), default=15)
        hwruntime_args.add_argument('--max_deps_per_task', help='maximum number of dependencies for any task in the bitstream\n(def: \'8\')', type=IntRangeType(imin=2), default=8)
        hwruntime_args.add_argument('--max_copies_per_task', help='maximum number of copies for any task in the bitstream\n(def: \'15\')', type=IntRangeType(imin=1), default=15)

        # Picos arguments
        picos_args = self.parser.add_argument_group('Picos')
        picos_args.add_argument('--picos_num_dcts', help='number of DCTs instantiated\n(def: \'1\')', choices=['1', '2', '4'], metavar='NUM_DCTS', default=1)
        picos_args.add_argument('--picos_tm_size', help='size of the TM memory\n(def: \'128\')', type=IntRangeType(imin=2), default=128)
        picos_args.add_argument('--picos_dm_size', help='size of the DM memory\n(def: \'512\')', type=IntRangeType(imin=2), default=512)
        picos_args.add_argument('--picos_vm_size', help='size of the VM memory\n(def: \'512\')', type=IntRangeType(imin=2), default=512)
        picos_args.add_argument('--picos_dm_ds', help='data structure of the DM memory\nBINTREE: Binary search tree (not autobalanced)\nLINKEDLIST: Linked list\n(def: \'BINTREE\')', choices=['BINTREE', 'LINKEDLIST'], metavar='DATA_STRUCT', default='BINTREE')
        picos_args.add_argument('--picos_dm_hash', help='hashing function applied to dependence addresses\nP_PEARSON: Parallel Pearson function\nXOR\n(def: \'P_PEARSON\')', choices=['P_PEARSON', 'XOR'], metavar='HASH_FUN', default='P_PEARSON')  # TODO: Explain what XOR does
        picos_args.add_argument('--picos_hash_t_size', help='DCT hash table size\n(def: \'64\')', type=IntRangeType(imin=2), default=64)

        # User-defined files arguments
        user_args = self.parser.add_argument_group('User-defined files')
        user_args.add_argument('--user_constraints', help='path of user defined constraints file', type=FileType())
        user_args.add_argument('--user_pre_design', help='path of user TCL script to be executed before the design step (not after the board base design)', type=FileType())
        user_args.add_argument('--user_post_design', help='path of user TCL script to be executed after the design step', type=FileType())

        # Miscellaneous arguments
        misc_args = self.parser.add_argument_group('Miscellaneous')
        misc_args.add_argument('-h', '--help', action='help', help='show this help message and exit')
        misc_args.add_argument('-i', '--verbose_info', help='print extra information messages', action='store_true', default=False)
        misc_args.add_argument('-j', '--jobs', help='specify the number of jobs to run simultaneously\nBy default it will use as many jobs as cores with at least 3GB of dedicated free memory, or the value returned by `nproc`, whichever is less.', type=IntRangeType(imin=1), default=getNumJobs())
        misc_args.add_argument('-k', '--keep_files', help='keep files on error', action='store_true', default=False)
        misc_args.add_argument('-v', '--verbose', help='print vendor backend messages', action='store_true', default=False)
        misc_args.add_argument('--version', help='print AIT version and exits', action='version', version=str(LONG_VERSION))

        # Hidden arguments
        hidden_args = self.parser.add_argument_group('Hidden')
        hidden_args.add_argument('--task_creation', help=argparse.SUPPRESS, action='store_true', default=False)
        hidden_args.add_argument('--deps_hwruntime', help=argparse.SUPPRESS, action='store_true', default=False)
        hidden_args.add_argument('--lock_hwruntime', help=argparse.SUPPRESS, action='store_true', default=False)

    def parse_args(self):
        # Check if configuration file exists and parse its values
        config_file_path = os.path.expanduser('~') + '/.ait/config.json'
        if os.path.exists(config_file_path):
            with open(config_file_path) as config_file:
                try:
                    config = json.load(config_file)
                    self.parser.set_defaults(**config)
                except ValueError as err:
                    print('Invalid configuration file (' + config_file_path + '). ' + str(err))
                    print('================')
                    config_file.seek(0)
                    print(config_file.read())
                    print('================')
                    sys.exit(1)

        # Parse values from argv
        args, extras = self.parser.parse_known_args()

        # Get vendor backend and parse vendor-specific arguments
        args_vars = vars(args)
        args_vars['backend'] = [backend for backend in backends if args.board in backends[backend]['boards']][0]
        if (args.backend in self.backend_parser):
            args, extras = self.backend_parser[args.backend].parser.parse_known_args(extras, namespace=args)
        if len(extras):
            self.parser.error('unrecognized arguments: ' + ','.join(extras) + '. Try \'ait -h\' for more information.')

        # Set default values for non-provided options
        for arg in self.defaults[args_vars['backend']]:
            if arg not in args_vars or args_vars[arg] is None:
                args_vars[arg] = self.defaults[args_vars['backend']][arg]['value']
                self.defaults[args_vars['backend']][arg]['used'] = True

        self.check_args(args)

        return args

    def check_args(self, args):
        # Validate arguments
        self.check_required_args(args)
        self.check_flow_args(args)
        self.check_bitstream_args(args)
        self.check_misc_args(args)

    def check_required_args(self, args):
        # Validate required args
        if not re.match('^[A-Za-z][A-Za-z0-9_]*$', args.name):
            msg.error('Invalid project name. Must start with a letter and contain only letters, numbers or underscores')

        if args.wrapper_version and args.wrapper_version < MIN_WRAPPER_VERSION:
            msg.error('Unsupported wrapper version (' + str(args.wrapper_version) + '). Minimum version is ' + str(MIN_WRAPPER_VERSION))

        if not os.path.exists(args.dir + '/' + args.name + '_ait'):
            os.mkdir(args.dir + '/' + args.name + '_ait')

    def check_flow_args(self, args):
        # Validate flow args
        if args.from_step not in backends[args.backend]['steps']:
            msg.error('Initial step \'' + args.from_step + '\' is not a valid generation step for \'' + args.backend + '\' backend. Set it correctly')

        if args.to_step not in backends[args.backend]['steps']:
            msg.error('Final step \'' + args.to_step + '\' is not a valid generation step for \'' + args.backend + '\' backend. Set it correctly')

        if backends[args.backend]['steps'].index(args.from_step) > backends[args.backend]['steps'].index(args.to_step):
            msg.error('Initial step \'' + args.from_step + '\' is posterior to the final step \'' + args.to_step + '\'. Set them correctly')

        if not args.disable_IP_caching and not os.path.isdir(args.IP_cache_location):
            if self.is_default('IP_cache_location', args.backend):
                # Create cache folder and set perms to allow all users writing there
                os.makedirs(args.IP_cache_location)
                os.chmod(args.IP_cache_location, 0o777)
            else:
                msg.error('Cache location (' + args.IP_cache_location + ') does not exist or is not a folder')

    def check_bitstream_args(self, args):
        # Validate bitstream args
        if len(args.bitinfo_note) > 256:
            msg.error('Length of bitInfo note must be less than 256 ASCII chars')

    # This check has to be delayed because arguments are parsed before the number of accelerators is calculated
    def check_hardware_runtime_args(self, args, num_accs):
        # Validate hardware runtime args
        def prev_power_of_2(num):
            if num & (num - 1) != 0:
                num = int(math.log2(num))
                num = int(pow(2, num))
            return num

        if args.cmdin_subqueue_len is not None and args.cmdin_queue_len is not None:
            msg.error('--cmdin_subqueue_len and --cmdin_queue_len are mutually exclusive')
        if args.cmdout_subqueue_len is not None and args.cmdout_queue_len is not None:
            msg.error('--cmdout_subqueue_len and --cmdout_queue_len are mutually exclusive')

        if args.cmdin_queue_len is not None:
            args.cmdin_subqueue_len = prev_power_of_2(int(args.cmdin_queue_len / num_accs))
            msg.info('Setting --cmdin_subqueue_len to {}'.format(args.cmdin_subqueue_len))
        elif args.cmdin_subqueue_len is None:
            args.cmdin_subqueue_len = max(64, prev_power_of_2(int(1024 / num_accs)))
        if args.cmdout_queue_len is not None:
            args.cmdout_subqueue_len = prev_power_of_2(int(args.cmdout_queue_len / num_accs))
            msg.info('Setting --cmdout_subqueue_len to {}'.format(args.cmdout_subqueue_len))
        elif args.cmdout_subqueue_len is None:
            args.cmdout_subqueue_len = max(64, prev_power_of_2(int(1024 / num_accs)))

        # The subqueue length has to be checked here in the case the user provides the cmdin queue length
        if args.cmdin_subqueue_len < 34:
            msg.warning('Value of --cmdin_subqueue_len={} is less than 34, which is the length of the longest command possible. This design might not work with tasks with enough arguments.'.format(args.cmdin_subqueue_len))
        if not args.disable_spawn_queues and args.spawnout_queue_len < 79:
            msg.warning('Value of --spawnout_queue_len={} is less than 79, which is the length of the longest task possible. This design might not work if an accelerator creates SMP tasks with enough copies, dependencies and/or arguments.'.format(args.spawnout_queue_len))

    def check_picos_args(self, args):
        # Validate Picos args
        if (args.picos_dm_hash == 'P_PEARSON' and args.picos_hash_t_size != 64):
            msg.error('With P_PEARSON hash function, --picos_hash_t_size must be 64')
        if (args.picos_hash_t_size > args.picos_dm_size):
            msg.error('Invalid --picos_hash_t_size ({}), maximum value is --picos_dm_size ({})'.format(args.picos_hash_t_size, args.picos_dm_size))
        if (math.ceil(math.log2(args.picos_hash_t_size)) + math.ceil(math.log2(args.picos_num_dcts)) > 8):
            msg.error('Invalid combination of --picos_hash_t_size and --picos_num_dcts, math.ceil(math.log2(args.picos_hash_t_size))+math.ceil(math.log2(args.picos_num_dcts)) <= 8')

    def check_misc_args(self, args):
        if args.jobs > getNumJobs():
            msg.warning('Using more Vivado jobs ({}) than the recommended default ({}). Performance of the compilation process might be affected'.format(args.jobs, getNumJobs()))

    def is_default(self, dest, backend):
        value = False
        if self.defaults[backend][dest] is not None:
            value = self.defaults[backend][dest]['used']
        return value



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

MIN_PYTHON_VERSION = (3, 7)
MIN_WRAPPER_VERSION = 8
BITINFO_VERSION = 9
VERSION_MAJOR = 6
VERSION_MINOR = 4
VERSION_PATCH = 0

# NOTE: The variable will be overwritten during installation, do not manually modify.
VERSION_COMMIT = 'commit: d7770a2'

SHORT_VERSION = str('.'.join([str(VERSION_MAJOR), str(VERSION_MINOR), str(VERSION_PATCH)]))  # Short numerical version
LONG_VERSION = str(SHORT_VERSION + ' (' + ', '.join(filter(None, [VERSION_COMMIT, 'bitinfo: ' + str(BITINFO_VERSION)])) + ')')  # Long version including commit hash/release tag and bitinfo version



#!/usr/bin/env python3
#
# ------------------------------------------------------------------------ #
#     (C) Copyright 2017-2022 Barcelona Supercomputing Center              #
#                             Centro Nacional de Supercomputacion          #
#                                                                          #
#     This file is part of OmpSs@FPGA toolchain.                           #
#                                                                          #
#     This code is free software; you can redistribute it and/or modify    #
#     it under the terms of the GNU Lesser General Public License as       #
#     published by the Free Software Foundation; either version 3 of       #
#     the License, or (at your option) any later version.                  #
#                                                                          #
#     OmpSs@FPGA toolchain is distributed in the hope that it will be      #
#     useful, but WITHOUT ANY WARRANTY; without even the implied           #
#     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.     #
#     See the GNU Lesser General Public License for more details.          #
#                                                                          #
#     You should have received a copy of the GNU Lesser General Public     #
#     License along with this code. If not, see <www.gnu.org/licenses/>.   #
# ------------------------------------------------------------------------ #

import glob
import importlib
import json
import os
import re
import subprocess
import sys
import time

from ait.frontend.parser import ArgParser
from ait.frontend.utils import Accelerator, ait_path, backends, msg, secondsToHumanReadable


class Logger(object):
    def __init__(self, project_path):
        self.terminal = sys.stdout
        self.log = open(project_path + '/' + args.name + '.ait.log', 'w+')
        self.subprocess = subprocess.PIPE if args.verbose else self.log
        self.re_color = re.compile(r'\033\[[0,1][0-9,;]*m')
        self.tag = '[AIT] ' if not args.verbose else ''

    def write(self, message):
        self.terminal.write(message)
        if message != '\n':
            self.log.write(self.tag)
        self.log.write(self.re_color.sub('', message))
        self.log.flush()

    def flush(self):
        pass


def check_board_support(board):
    chip_part = board.chip_part + ('-' + board.es if board.es and not args.ignore_eng_sample else '')

    if args.verbose_info:
        msg.log('Checking vendor support for selected board')

    backend_utils = importlib.import_module('ait.backend.{}.utils'.format(args.backend))
    checker = getattr(backend_utils, 'checkers')
    checker.check_board_support(chip_part)


def get_accelerators(project_path):
    global accs

    if args.verbose_info:
        msg.log('Searching accelerators in folder: ' + os.getcwd())

    accs = []
    acc_types = []
    acc_names = []
    args.num_accs = 0
    args.num_instances = 0
    args.num_acc_creators = 0

    for file_ in sorted(glob.glob(os.getcwd() + '/ait_*.json')):
        acc_config_json = json.load(open(file_))
        for acc_config in acc_config_json:
            acc = Accelerator(acc_config)

            if not re.match('^[A-Za-z][A-Za-z0-9_]*$', acc.name):
                msg.error('\'' + acc.name + '\' is an invalid accelerator name. Must start with a letter and contain only letters, numbers or underscores')

            msg.info('Found accelerator \'' + acc.name + '\'')

            args.num_accs += 1
            args.num_instances += acc.num_instances

            if acc.type in acc_types:
                msg.error('Two accelerators use the same type: \'' + str(acc.type) + '\' (maybe you should use the onto clause)')
            elif acc.name in acc_names:
                msg.error('Two accelerators use the same name: \'' + str(acc.name) + '\' (maybe you should change the fpga task definition)')
            acc_types.append(acc.type)
            acc_names.append(acc.name)

            # Check if the acc is a task creator
            if acc.task_creation:
                args.task_creation = True
                args.num_acc_creators += acc.num_instances
                accs.insert(0, acc)
            else:
                accs.append(acc)

            # Check if the acc needs instrumentation support
            if acc.instrumentation:
                args.hwinst = True

            # Check if the acc needs lock support
            if acc.lock:
                args.lock_hwruntime = True

            # Check if the acc needs dependencies
            if acc.deps:
                args.deps_hwruntime = True

    if args.num_accs == 0:
        msg.error('No accelerators found')
    elif args.num_acc_creators == 0:
        args.disable_spawn_queues = True
        args.spawnin_queue_len = 0
        args.spawnout_queue_len = 0

    # Generate the .xtasks.config file
    xtasks_config_file = open(project_path + '/' + args.name + '.xtasks.config', 'w')
    xtasks_config = 'type\t#ins\tname\t    \n'
    for acc in accs:
        xtasks_config += str(acc.type).zfill(19) + '\t' + str(acc.num_instances).zfill(3) + '\t' + acc.name.ljust(31)[:31] + '\t000\n'
    xtasks_config_file.write(xtasks_config)
    xtasks_config_file.close()

    if args.hwinst:
        hwinst_acc_json_string = json.dumps({'full_path': ait_path + '/backend/' + args.backend + '/HLS/src/Adapter_instr.cpp', 'filename': 'Adapter_instr.cpp', 'name': 'Adapter_instr', 'type': 0, 'num_instances': 1, 'task_creation': 'false', 'instrumentation': 'false', 'periodic': 'false', 'lock': 'false'}, indent=4)
        hwinst_acc_json = json.loads(hwinst_acc_json_string)
        hwinst_acc = Accelerator(hwinst_acc_json)
        accs.append(hwinst_acc)


def main():
    global args
    global backend

    start_time = time.time()

    parser = ArgParser()
    args = parser.parse_args()

    msg.setProjectName(args.name)
    msg.setPrintTime(args.verbose_info)
    msg.setVerbose(args.verbose)

    msg.info('Using ' + args.backend + ' backend')

    driver = importlib.import_module('ait.backend.{}.driver'.format(args.backend))
    steps, board = driver.load(args)

    project_path = os.path.normpath(os.path.realpath(args.dir + '/' + args.name + '_ait'))

    sys.stdout = Logger(project_path)
    sys.stdout.log.write(os.path.basename(sys.argv[0]) + ' ' + ' '.join(sys.argv[1:]) + '\n\n')

    get_accelerators(project_path)

    parser.check_hardware_runtime_args(args, max(2, args.num_instances))
    if args.deps_hwruntime:
        parser.check_picos_args(args)

    project_args = {
        'path': os.path.normpath(os.path.realpath(args.dir) + '/' + args.name + '_ait'),
        'accs': accs,
        'board': board,
        'args': args
    }

    for step in backends[args.backend]['steps']:
        if backends[args.backend]['steps'].index(args.from_step) <= backends[args.backend]['steps'].index(step) <= backends[args.backend]['steps'].index(args.to_step):
            step_func = getattr(steps, step)
            msg.info('Starting \'' + step + '\' step')
            step_start_time = time.time()
            project_args['start_time'] = step_start_time
            step_func.run_step(project_args)
            msg.success('Step \'' + step + '\' finished. ' + secondsToHumanReadable(int(time.time() - step_start_time)) + ' elapsed')
        else:
            msg.warning('Step \'' + step + '\' is disabled')

    msg.success('Accelerator automatic integration finished. ' + secondsToHumanReadable(int(time.time() - start_time)) + ' elapsed')


if __name__ == '__main__':
    main()



import subprocess

import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        with open('ait/frontend/config.py', 'r') as config_file:
            config = config_file.read()

        config_commit = config.replace("VERSION_COMMIT = '{}'".format(version_commit), "VERSION_COMMIT = ''")

        with open('ait/frontend/config.py', 'w') as config_file:
            config_file.write(config_commit)


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        with open('ait/frontend/config.py', 'r') as config_file:
            config = config_file.read()

        config_commit = config.replace("VERSION_COMMIT = '{}'".format(version_commit), "VERSION_COMMIT = ''")

        with open('ait/frontend/config.py', 'w') as config_file:
            config_file.write(config_commit)


tag = str(subprocess.check_output(['git describe --tags --exact-match HEAD 2>/dev/null || true'], shell=True), 'utf-8').strip()
commit_hash = str(subprocess.check_output(['git show -s --format=%h 2>/dev/null || true'], shell=True), 'utf-8').strip()

version_commit = ''
if tag:
    version_commit = tag
elif commit_hash:
    commit_hash += str(subprocess.check_output(["git diff --quiet HEAD || echo '-dirty'"], shell=True), 'utf-8').strip()
    version_commit = 'commit: ' + commit_hash

if version_commit:
    with open('ait/frontend/config.py', 'r') as config_file:
        config = config_file.read()

    config_commit = config.replace("VERSION_COMMIT = ''", "VERSION_COMMIT = '{}'".format(version_commit))

    with open('ait/frontend/config.py', 'w') as config_file:
        config_file.write(config_commit)

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    }
)

if version_commit:
    with open("ait/frontend/config.py", "w") as config_file:
        config_file.write(config)
